// requires AE.Translate, AE.Bridge
// TODO: for editable commands, show small toolbar for edit and remove


var AE = window.AE || {};


/**
 * @module
 */
AE.ToolbarEditor = function(self) {


// Object to hold data.
var Index = [];
var availableButtons = [];
var toolbars = [];


/**
 * Escapes a String for generating a regular Expression.
 * @param {string} string
 * @returns {string}
 * @private
 */
var escapeRegExp = function(string) {
  return string.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
}


/**
 * jQuery on document ready
 */
$(function() {

  // Allow to remove items from sortable.
  $("#buttons-available").droppable({
    accept: ".toolbar .toolbar-button",
    drop: function( event, ui ) {
      ui.helper.remove();
      ui.draggable.remove();
    }
  });

  // Button to add a new toolbar
  $("#button-add-toolbar").on("click", function() { new Toolbar().activate() });

  // Search box // TODO: Does this sometimes not trigger in IE?
  $("#input-filter").on("keydown", function() {
    var match;
    var searchWord = $("#input-filter").val();
    if (searchWord.length > 0) {
      $(availableButtons).hide();
      var searchExp = new RegExp(escapeRegExp(searchWord), "i");
      $("#buttons-available li").each(function() {
        match = ($(this).text().search(searchExp) !== -1);
        if (match) { $(this).show(); }
      });
    }
    // If no filter is applied, show all again.
    else {
      $(availableButtons).show();
    }
  });

  // Button to add a new custom button
  $("#button-add-custom-button").on("click", function() { 
    self.showDialogEditCustomButton(function(entry) {
      var button = new CustomButton(entry);
      // Allow to drag a copy of the button into any toolbars.
      button.jQueryElement.draggable({
        connectToSortable: ".toolbar",
        // We need to put the helper outside of the overflow:scroll/hidden element.
        helper: "clone",
        appendTo: document.body,
        scroll: false,
        revert: "invalid",
        cursor: "move", cursorAt: { top: 10, left: 20 }
      });
      // Add the button to the list.
      availableButtons = availableButtons.push(button.jQueryElement.get(0));
      $("#buttons-available").prepend(button.jQueryElement.get(0));
    });
  });

  // Apply button
  $("#button-apply").on("click", function() {
    var data = [], toolbar;
    for (var i = 0; i < toolbars.length; i++) {
      toolbar = toolbars[i];
      if (toolbar.buttons.length > 0) {
        data.push({"name": toolbar.name, "id": toolbar.id, "buttons": toolbar.buttons});
      };
    };
    AE.Bridge.callRuby("save_toolbars", data);
  });

  // Cancel button
  $("#button-cancel").on("click", function() {
    AE.Bridge.callRuby("close");
  });

  // Disable text selection on various UI elements.
  $("ul, li").disableSelection();

});


/**
 * Build a list of custom buttons.
 * @param {array<object>} array - list of objects <name,description,icon,id>
 */
self.loadCustomButtons = function(array) {
  if (!Object.prototype.toString.call(array) === '[object Array]') { return; }
  // Keep the button metadata in an index.
  Index = Index.concat(array);
  // Generate buttons.
  var buttons = [], button, entry;
  for (var i = 0; i < array.length; i++) {
    entry = array[i];
    button = new CustomButton(entry);
    if (!button.jQueryElement) { continue; }
    // Allow to drag a copy of the button into any toolbars.
    button.jQueryElement.draggable({
      connectToSortable: ".toolbar",
      // We need to put the helper outside of the overflow:scroll/hidden element.
      helper: "clone",
      appendTo: document.body,
      scroll: false,
      revert: "invalid",
      cursor: "move", cursorAt: { top: 10, left: 20 }
    });
    buttons.push(button.jQueryElement.get(0));
  }
  // Add the buttons to the list.
  availableButtons = availableButtons.concat(buttons);
  $("#buttons-available").prepend(buttons);
};


/**
 * Build a list of available buttons.
 * @param {array<object>} array - list of objects <name,description,icon,id>
 */
self.loadButtons = function(array) {
  if (!Object.prototype.toString.call(array) === '[object Array]') { return; }
  // Loading of buttons finished, now we can remove the indicator.
  $("#loading").remove();
  // Add a separator.
  array.unshift({
    id: 0,
    name: AE.Translate.get("Separator"),
    description: AE.Translate.get("Add a separator to group toolbar buttons."),
    icon: "../images/separator.png"
  });
  // Keep the button metadata in an index.
  Index = Index.concat(array);
  // Generate buttons.
  var buttons = [], button, entry;
  for (var i = 0; i < array.length; i++) {
    entry = array[i];
    button = new ToolbarButton(entry);
    if (!button.jQueryElement) { continue; }
    // Allow to drag a copy of the button into any toolbars.
    button.jQueryElement.draggable({
      connectToSortable: ".toolbar",
      // We need to put the helper outside of the overflow:scroll/hidden element.
      helper: "clone",
      appendTo: document.body,
      scroll: false,
      revert: "invalid",
      cursor: "move", cursorAt: { top: 10, left: 20 }
    });
    buttons.push(button.jQueryElement.get(0));
  }
  // Add the buttons to the list.
  availableButtons = availableButtons.concat(buttons);
  $("#buttons-available").append(buttons);
};


/**
 * Hide all buttons except those with the given ids.
 * @param {array<number>} ids 
 */
self.filterButtons = function(ids) { // TODO: currently not used
  if (ids && ids.length > 0) {
    $(availableButtons).hide();
    for (var i = 0; i < ids.length; i++) {
      $("#" + ids[i]).show();
    }
  }
  // If no filter is applied, show all again.
  else {
    $(availableButtons).show();
  }
};


/**
 * Shows the dialog for editing a custom button.
 * @param {object} entry
 */
self.showDialogEditCustomButton = function(callback, entry) {
  if (!entry) { entry = {}; } // new button
  $("#edit-custom-button-name").val(entry.name); //.attr("disabled", !!entry.id);
  $("#edit-custom-button-category").val(entry.category);
  $("#edit-custom-button-description").val(entry.description);
  $("#edit-custom-button-icon").val(entry.icon);
  $("#edit-custom-button-image").attr("src", entry.icon);
  $("#edit-custom-button-code").val(entry.code);
  $("#dialog-edit-custom-button").show();
  // Cancel button
  $("#button-edit-custom-cancel").on("click", function() {
    $("#dialog-edit-custom-button").hide();
    $("#button-edit-custom-cancel").off("click");
    $("#button-edit-custom-save").off("click");
  });
  // Save button
  $("#button-edit-custom-save").on("click", function() {
    entry.name = $("#edit-custom-button-name").val();
    if (!entry.id) { entry.id = hashCode(entry.name + entry.description + entry.icon); }
    entry.category = $("#edit-custom-button-category").val();
    entry.description = $("#edit-custom-button-description").val();
    entry.icon = $("#edit-custom-button-icon").val();
    entry.code = $("#edit-custom-button-code").val();
    // Save entry to SketchUp
    AE.Bridge.callRuby("save_snippet", entry.id, entry.code, 
      {name: entry.name, category: entry.category, description: entry.description, icon: entry.icon},
      function(success) {
        if (success) {
          $("#dialog-edit-custom-button").hide();
          $("#button-edit-custom-cancel").off("click");
          $("#button-edit-custom-save").off("click");
        } else {
          // Show warning about error
          $("#edit-custom-button-code").css({border: "3px red solid"});
          setTimeout(function(){ 
            $("#edit-custom-button-code").css({border: ""});
          }, 500);
        }
    });
    if (typeof callback === "function") { callback(entry); }
  });
};


/**
 * Hash function to get short and reproducible identifiers.
 * @param {object} object
 */
function adler32(s, l) {
  var s = String(s),
    h = 0,
    m = 99999,
    c;
  for (var i = 0; i < s.length; i++) {
    c = s.charCodeAt(i);
    h = (128 * h + c) % m;
  }
  return h;
}
var hashCode = adler32;


/**
 * Build a list of created toolbars.
 * @param {array<object>} array - list of objects <name,description,icon,id>
 */
self.loadToolbars = function(array) {
  if (!Object.prototype.toString.call(array) === '[object Array]') { return; }
  var toolbar, buttons;
  for (var i=0; i<array.length; i++) {
    if (!array[i]) { continue; };
    if (array[i]["id"] >= toolbarCounter) { toolbarCounter = array[i]["id"] + 1; }
    toolbar = new Toolbar(array[i]["name"]);
    toolbar.id = array[i]["id"];
    buttons = array[i]["buttons"];
    for (var j=0; j<buttons.length; j++) {
      toolbar.addButton(buttons[j]);
    }
  }
  // Create a new toolbar if none exists.
  if (array.length === 0) {
    toolbar = new Toolbar();
  }
  // Select the last toolbar.
  toolbars[toolbars.length-1].activate();
};


/**
 * @class Toolbar
 * @param {string} name
 */
var toolbarCounter = 0;
var Toolbar = function(name) {

  /** @private */
  var toolbar = this;
  var header, title, titleEdit, buttonRemove, buttonEdit, buttonWrapper, content, statusEditName = false;

  /** @public */
  toolbar.id = (toolbarCounter += 1);
  toolbar.name = (typeof name == "string") ? name :
    AE.Translate.get("Toolbar") + " " + toolbar.id;
  toolbar.buttons = [];

  /**
   * Initializes the new toolbar and adds it to the UI.
   * @param {string} name
   * @constructor
   * @private
   */
  var initialize = function(name) {
    // Header with buttons.
    header = $("<h3/>").
      attr({"class": "toolbar header"});
    header.get(0).toolbar = toolbar; // reference
    title = $("<span/>").text(toolbar.name);
    titleEdit = $("<input/>").
      attr({"type": "text", "value": toolbar.name}).
      hide().
      on("keydown", function(event) {
        // Enter
        if (event.which == 13) {
          toolbar.finishEditName();
          event.preventDefault();
        }
        event.stopPropagation();
      }).
      on("blur", function() {
        toolbar.finishEditName();
        statusEditName = false;
      });
    // Button to remove the toolbar.
    buttonRemove = $("<button/>").
      attr({"class": "symbolic-button", "title": AE.Translate.get("Remove this toolbar")}).
      append($("<img/>").attr("src", "../images/icon_close.png")).
      on("click", function() {
        toolbar.remove();
      });
    // Button to edit a toolbar's name.
    buttonEdit = $("<button/>").
      attr({"class": "symbolic-button", "title": AE.Translate.get("Edit the toolbar name")}).
      append($("<img/>").attr("src", "../images/icon_edit.png")).
      on("click", function() {
        !statusEditName ? toolbar.editName() : toolbar.finishEditName();
        statusEditName = !statusEditName;
      });
    buttonWrapper = $("<span/>").
      attr({"class": "right"}).
      append(buttonEdit, buttonRemove);
    header.append(title, titleEdit, buttonWrapper, $("<div/>").attr({"class": "clear"}) );

    //Content
    content = $("<ul/>").
      attr({"class": "toolbar content"});
    content.get(0).toolbar = toolbar; // reference

    // Add to toolbars list.
    $("#custom-toolbars").append(header, content);

    // Only after the content has been added to the DOM, we can make it sortable.
    content.sortable({
      connectWith: ".toolbar",
      revert: true,
      // Refresh the toolbar buttons array:
      change: function(event, ui) { update() },
      stop: function(event, ui) { update() }, // Also triggered when items on droppable.
      scroll: false,
      appendTo: document.body
      //beforeStop: function(event, ui) { alert("beforeStop") }
      // Check that the dragged element is allowed to be dropped.
      /*beforeStop: function(event, ui) {
        // Check that it is no duplicate (except separator). When SketchUp loads toolbars, duplicates show a wrong icon.
        var id = Number(ui.item.attr("name"));
        if ( $.inArray(id, toolbar.buttons) >= 0 && id !== 0) { ui.item.remove(); }
      },*/
    });
  };

  /**
   * Refreshes the list of buttons of this toolbar.
   * @private
   */
  var update = function() {
    toolbar.buttons = content.find(".toolbar-button").map(function() {
      return Number(this.getAttribute("name"))
    }).get();
  };

  /**
   * Select the panel of this toolbar.
   * @public
   */
  toolbar.activate = function() {
    $("#custom-toolbars").parent().scrollTop(header.offset().top);
  };

  /**
   * Edit mode of this toolbar's name.
   */
  toolbar.editName = function() {
    titleEdit.show().focus();
    title.hide();
    /* Set caret to end */
    var e = titleEdit.get(0);
    var pos = e.value.length;
    /* DOM 3.0 */
    if('selectionStart' in e){
      e.selectionStart = pos;
      e.selectionEnd = pos;
    }
    /* Internet Explorer */
    else if (document.selection) {
      var r = e.createTextRange();
      r.moveEnd('textedit', -1);
      r.moveStart('character', pos);
      r.moveEnd('character', 0);
      r.select();
    }
  };

  /**
   * Finish edit mode of this toolbar's name.
   */
  toolbar.finishEditName = function() {
    if (statusEditName) {
      toolbar.name = titleEdit.hide().val();
      title.text(toolbar.name).show();
    }
  };

  /**
   * Add a new button (programmatically, not drag&drop)
   * @param {object,number} hash_or_id
   * @public
   */
  toolbar.addButton = function(hash_or_id) {
    var button = new ToolbarButton(hash_or_id);
    if (!button) { return; }
    content.append(button.jQueryElement);
    toolbar.buttons.push(button.id);
  };

  /**
   * Remove this toolbar completely.
   * @public
   */
  toolbar.remove = function() {
    header.remove();
    content.remove();
    toolbars.splice($.inArray(toolbar, toolbars), 1);
    delete toolbar;
  };

  initialize();
  toolbars.push(toolbar);
}; // end @class Toolbar


/**
 * @class ToolbarButton
 */
var ToolbarButton = function(id_or_hash) {

  /** @private */
  var button = this;
  if (Object.prototype.toString.call(id_or_hash) === '[object Object]') { // hash
    var entry = id_or_hash;
    button.id = entry.id;
  } else { // id
    var entry = $.grep(Index, function(item){
      return item.id == id_or_hash;
    })[0];
    button.id = id_or_hash;
  };
  if (!entry || typeof entry.id !== "number") { return; }

  /**
   * @constructor
   */
  var initialize = function() {
    button.jQueryElement = $("<li/>").
      attr({"id": entry.id, "name": entry.id, "title": entry.name, "class": "toolbar-button"}).
      data("entry", entry).
      append(
        $("<img/>").attr({"class": "icon", "src": entry.icon}),
        $("<span/>").attr({"class": "name"}).text(entry.name),
        $("<span/>").attr({"class": "category"}).text(entry.category),
        $("<span/>").attr({"class": "description"}).text(entry.description)
      );
  };

  initialize();
}; // end @class ToolbarButton


/**
 * @class CustomButton
 */
var CustomButton = function(entry) {

  /** @private */
  var button = this;
  button.id = entry.id;
  if (!entry || !button.id) { return; }
  var image_element, name_element, category_element, description_element;

  /**
   * @constructor
   */
  var initialize = function() {
    button.jQueryElement = $("<li/>").
      attr({"id": entry.id, "name": entry.id, "title": entry.name, "class": "toolbar-button"}).
      append(
        $("<div/>").attr("class", "controls").append(
          // Delete button
          $("<div/>").append(
            $("<button/>").attr("title", AE.Translate.get("Delete")).addClass("symbolic-button").on("click", function() {
              var index = 0; while (availableButtons[index] && availableButtons[index] != button) { index++; }
              availableButtons.splice(index, 1);
              button.jQueryElement.remove();
              AE.Bridge.callRuby("delete_snippet", entry.id);
            }).append(
              $("<img/>").attr({"src": "../images/icon_close.png"})
            )
          ),
          // Edit button
          $("<div/>").append(
            $("<button/>").attr("title", AE.Translate.get("Edit")).addClass("symbolic-button").on("click", function() {
              self.showDialogEditCustomButton(function(entry) {
                // Update the custom button.
                image_element.attr("src", entry.icon);
                name_element.text(entry.name);
                category_element.text(entry.category);
                description_element.text(entry.description);
              }, entry);
            }).append(
              $("<img/>").attr({"src": "../images/icon_edit.png"})
            )
          )
        ),
        image_element = $("<img/>").attr({"class": "icon", "src": entry.icon}),
        name_element = $("<span/>").attr({"class": "name"}).text(entry.name),
        category_element = $("<span/>").attr({"class": "category"}).text(entry.category),
        description_element = $("<span/>").attr({"class": "description"}).text(entry.description)
      );
  };


  initialize();
}; // end @class CustomButton*/


return self;
}(AE.ToolbarEditor || {}); // end @module ToolbarEditor


window.onerror = function(message, url, lineNumber){
  AE.Bridge.error(message+" \n"+url+"\n"+lineNumber);
  return true;
}
